//*****************************************************************************
//* BootLoader 4.0 for MegaLoad 4.0
//* Modified for avr-gcc by Werner B.
//*
//* Devices supported at this time
//* ATMega8
//* ATMega16
//* ATMega32
//* ATMega64
//* ATMega128
//* ATMega162
//* ATMega169
//* ATMega8515
//* ATMega8535
//* ATMega48    not tested, please give me feedback!
//* ATMega88    not tested, please give me feedback!
//* ATMega168   not tested, please give me feedback!
//* ATMega165   not tested, please give me feedback!
//* ATMega3250  not tested, please give me feedback!
//* ATMega6450  not tested, please give me feedback!
//* ATMega168   not tested, please give me feedback!
//* ATMega3290  not tested, please give me feedback!
//* ATMega6490  not tested, please give me feedback!
//* ATMega406   not tested, please give me feedback!
//* ATMega640   not tested, please give me feedback!
//* ATMega1280  not tested, please give me feedback!
//* ATMega2560  not tested, please give me feedback!
//*
//* Version 4.0 June 2005
//* 1.0 : First Release
//* 1.1 : Add auto baud rate, code smaller
//* 1.2 : Correct a bug with file larger than 65280 bytes.
//* 2.0 : All in one source file that fit in 512 Words
//* 3.0 : Add EEprom programming
//*       Add RS485 half duplex support
//*       Add LockBit programming
//* 3.1 : Correct a problem with interupt vector for M8/16/32/162/8515/8535
//*       Change some #define to be more clear
//* 3.2 : Add support for M48/M88/M168
//* 4.0 : Change code to work with MegaLoad .NET 4.0
//*       Add M165/325x/645x/168/329x/649x/406/640/128x/256x
//*
//* Bissonnette.Sylvain@MicroSyl.com
//*
//*****************************************************************************
//
//
#include <avr/interrupt.h>
#include <avr/boot.h>

// Define for MegaLoad .NET v4.0x  -  undefine for MegaLoad v3
//#define MEGALOAD_NET  -- set in Makefile

//*****************************************************************************
// Auto OSCCAL
// Enable the autocalibration of OSCCAL register (can't be used with autobaud)
//
//*****************************************************************************
//#define AUTOOSCCAL

//*****************************************************************************
// RS485
// if you use RS485 half duplex for bootloader
// make the appropriate change for RX/TX transceiver switch
//*****************************************************************************
//#define RS485DDR  DDRB
//#define RS485PORT PORTB
//#define RS485TXE  0x08

//*****************************************************************************
// BaudRate
// If you don't specify the baudrate divisor the bootloader
// will automaticaly be in AutoBaudRate mode
// Preferred set in Makefile
//*****************************************************************************
//#if !defined (BAUDRATE)
//#define BAUDRATE     38400
//#endif

//*****************************************************************************
// Bootload on UART x
//*****************************************************************************
// defined in Makefile !
//#define UART        0
//#define UART        1
//#define UART        2
//#define UART        3

//*****************************************************************************
// EEprom programming
// enable EEprom programing via bootloader
//*****************************************************************************
// defined in Makefile !
//#define EEPROM

//*****************************************************************************
// LockBit programming
// enable LOCKBIT programing via bootloader
//*****************************************************************************
// defined in Makefile !
//#define LOCKBIT

#define  BootSize       Boot512

#if !defined(AUTOOSCCAL) && !defined(HOMENET) && !defined(RS485DDR)
#if defined(__AVR_ATmega8__)||defined(__AVR_ATmega8515__)||defined(__AVR_ATmega8535__)||defined(__AVR_ATmega88__)
  #undef  BootSize
  #define  BootSize       Boot256
#endif

#if defined(__AVR_ATmega16__)||defined(__AVR_ATmega162__)||defined(__AVR_ATmega48__)
 #if !defined(EEPROM)||(!defined(LOCKBIT)&&!defined(MEGALOAD_NET)&&defined(BAUDRATE))
  #undef  BootSize
  #define  BootSize       Boot256
 #endif
#endif

#if defined(__AVR_ATmega32__)
 #if !defined(EEPROM)||!defined(LOCKBIT)
  #undef  BootSize
  #define  BootSize       Boot256
 #endif
#endif

#if defined(__AVR_ATmega64__)
 #if !defined(EEPROM)
  #undef  BootSize
  #define  BootSize       Boot256
 #endif
#endif
#endif
/*
  MegaLoad.NET uses a different startup sequence ..
*/
#if defined (MEGALOAD_NET)
 #define START_CH  0x55
 #define START_CH_RET 0x55
#else
 #define START_CH  '>'
 #define START_CH_RET '<'
#endif

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//!!!!!!!!!!!!!! DO NOT CHANGE ANYTHING BELOW THIS LINE !!!!!!!!!!!!!!!!!!!!!!!
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#define Tiny2313        'O'
#define Tiny13          'P'
#define Mega8           'A'
#define Mega16          'B'
#define Mega32          'E'
#define Mega48          'L'
#define Mega64          'C'
#define Mega88          'M'
#define Mega128         'D'
#define Mega162         'F'
#define Mega163         'J'
#define Mega168         'N'
#define Mega169         'G'
#define Mega323         'K'
#define Mega8515        'H'
#define Mega8535        'I'
#define Mega165         0x80
#define Mega3250        0x81
#define Mega6450        0x82
#define Mega3290        0x83
#define Mega6490        0x84
#define Mega406         0x85
#define Mega640         0x86
#define Mega1280        0x87
#define Mega2560        0x88

#define Flash1k         'g'
#define Flash2k         'h'
#define Flash4k         'i'
#define Flash8k         'l'
#define Flash16k        'm'
#define Flash32k        'n'
#define Flash64k        'o'
#define Flash128k       'p'
#define Flash256k       'q'
#define Flash40k        'r'

#define EEprom64        '.'
#define EEprom128       '/'
#define EEprom256       '0'
#define EEprom512       '1'
#define EEprom1024      '2'
#define EEprom2048      '3'
#define EEprom4096      '4'

/* BootLoader size (WORDS) */
#define Boot128         'a'
#define Boot256         'b'
#define Boot512         'c'
#define Boot1024        'd'
#define Boot2048        'e'
#define Boot4096        'f'

#define Page32          'Q'
#define Page64          'R'
#define Page128         'S'
#define Page256         'T'
#define Page512         'V'

#if ((defined(__AVR_ATmega64__)||defined(__AVR_ATmega128__)||defined(__AVR_AT90CAN128__)||defined(__AVR_ATmega162__)||defined(__AVR_ATmega640__)||defined(__AVR_ATmega1280__)||defined(__AVR_ATmega2560__)) && !(defined UART))
  #error "UART ch must be defined for this part"
#endif

#if defined(__AVR_ATmega8__)
#define  DeviceID       Mega8
#define  FlashSize      Flash8k
#define  PageSize       Page64
#define  EEpromSize     EEprom512
#define  PageByte       64
#define  NSHIFTPAGE     6
#define  INTVECREG      GICR
#define  PULLUPPORT     PORTD
#define  PULLUPPIN      0x01
#endif

#if defined(__AVR_ATmega16__)
#define  DeviceID       Mega16
#define  FlashSize      Flash16k
#define  PageSize       Page128
#define  EEpromSize     EEprom512
#define  PageByte       128
#define  NSHIFTPAGE     7
#define  INTVECREG      GICR
#define  PULLUPPORT     PORTD
#define  PULLUPPIN      0x01
#endif

#if defined(__AVR_ATmega32__)
#define  DeviceID       Mega32
#define  FlashSize      Flash32k
#define  PageSize       Page128
#define  EEpromSize     EEprom1024
#define  PageByte       128
#define  NSHIFTPAGE     7
#define  INTVECREG      GICR
#define  PULLUPPORT     PORTD
#define  PULLUPPIN      0x01
#endif

#if defined(__AVR_ATmega64__)
#define  DeviceID       Mega64
#define  FlashSize      Flash64k
#define  PageSize       Page256
#define  EEpromSize     EEprom2048
#define  PageByte       256
#define  NSHIFTPAGE     8
#define  INTVECREG      MCUCR
#if (UART == 0)
#define PULLUPPORT      PORTE
#define PULLUPPIN       0x01
#endif

#if (UART == 1)
#define PULLUPPORT      PORTD
#define PULLUPPIN       0x04
#endif
#endif

#if defined(__AVR_ATmega128__)
#define  DeviceID       Mega128
#define  FlashSize      Flash128k
#define  PageSize       Page256
#define  EEpromSize     EEprom4096
#define  PageByte       256
#define  NSHIFTPAGE     8
#define  INTVECREG      MCUCR
#define  RAMPZ_FLAG
#if (UART == 0)
#define PULLUPPORT      PORTE
#define PULLUPPIN       0x01
#endif

#if (UART == 1)
#define PULLUPPORT      PORTD
#define PULLUPPIN       0x04
#endif
#endif

#if defined(__AVR_AT90CAN128__)
#define  DeviceID       Mega128
#define  FlashSize      Flash128k
#define  PageSize       Page256
#define  EEpromSize     EEprom4096
#define  PageByte       256
#define  NSHIFTPAGE     8
#define  INTVECREG      MCUCR
#define  RAMPZ_FLAG
#if (UART == 0)
#define PULLUPPORT      PORTE
#define PULLUPPIN       0x01
#endif

#if (UART == 1)
#define PULLUPPORT      PORTD
#define PULLUPPIN       0x04
#endif
#endif

#if defined(__AVR_ATmega162__)
#define  DeviceID       Mega162
#define  FlashSize      Flash16k
#define  PageSize       Page128
#define  EEpromSize     EEprom512
#define  PageByte       128
#define  NSHIFTPAGE     7
#define  INTVECREG      GICR
#if (UART == 0)
#define PULLUPPORT      PORTD
#define PULLUPPIN       0x01
#define FE         FE0
#define UPE        UPE0
#define RXC        RXC0
#endif

#if (UART == 1)
#define PULLUPPORT      PORTB
#define PULLUPPIN       0x04
#define FE         FE1
#define UPE        UPE1
#define RXC        RXC1
#endif
#endif

#if defined(__AVR_ATmega169__)
#define  DeviceID       Mega169
#define  FlashSize      Flash16k
#define  PageSize       Page128
#define  EEpromSize     EEprom512
#define  PageByte       128
#define  NSHIFTPAGE     7
#define  INTVECREG      MCUCR
#define  PULLUPPORT     PORTE
#define  PULLUPPIN      0x01
#define  PE             UPE
#endif

#if defined(__AVR_ATmega8515__)
#define  DeviceID       Mega8515
#define  FlashSize      Flash8k
#define  PageSize       Page64
#define  EEpromSize     EEprom512
#define  PageByte       64
#define  NSHIFTPAGE     6
#define  INTVECREG      GICR
#define  PULLUPPORT     PORTD
#define  PULLUPPIN      0x01
#endif

#if defined(__AVR_ATmega8535__)
#define  DeviceID       Mega8535
#define  FlashSize      Flash8k
#define  PageSize       Page64
#define  EEpromSize     EEprom512
#define  PageByte       64
#define  NSHIFTPAGE     6
#define  INTVECREG      GICR
#define  PULLUPPORT     PORTD
#define  PULLUPPIN      0x01
#endif

#if defined(__AVR_ATtiny2313__)
#define  DeviceID       Tiny2313
#define  FlashSize      Flash2k
#define  PageSize       Page32
#define  EEpromSize     EEprom128
#define  PageByte       32
#define  NSHIFTPAGE     5
#define  PULLUPPORT     PORTD
#define  PULLUPPIN      0x01
#endif

#if defined(__AVR_ATmega48__)
#define  DeviceID       Mega48
#define  FlashSize      Flash8k
#define  PageSize       Page128
#define  EEpromSize     EEprom256
#define  PageByte       128
#define  NSHIFTPAGE     7
#define  INTVECREG      MCUCR
#define  PULLUPPORT     PORTD
#define  PULLUPPIN      0x01
#ifdef UART
 #undef UART
#endif
#define UART       0
#endif

#if defined(__AVR_ATmega88__)
#define UART_NO
#define  DeviceID       Mega88
#define  FlashSize      Flash8k
#define  PageSize       Page128
#define  EEpromSize     EEprom512
#define  PageByte       128
#define  NSHIFTPAGE     7
#define  INTVECREG      MCUCR
#define  PULLUPPORT     PORTD
#define  PULLUPPIN      0x01
#ifdef UART
 #undef UART
#endif
#define UART       0
#endif

#if defined(__AVR_ATmega168__)
#define  DeviceID       Mega168
#define  FlashSize      Flash8k
#define  PageSize       Page128
#define  EEpromSize     EEprom512
#define  PageByte       128
#define  NSHIFTPAGE     7
#define  INTVECREG      MCUCR
#define  PULLUPPORT     PORTD
#define  PULLUPPIN      0x01
#ifdef UART
 #undef UART
#endif
#define UART       0
#endif

#if defined(__AVR_ATmega165__)
#define  DeviceID       Mega165
#define  FlashSize      Flash16k
#define  PageSize       Page128
#define  EEpromSize     EEprom512
#define  PageByte       128
#define  NSHIFTPAGE     7
#define  INTVECREG      MCUCR
#define  PULLUPPORT     PORTE
#define  PULLUPPIN      0x01
#endif

#if defined(__AVR_ATmega3250__)
#define  DeviceID       Mega3250
#define  FlashSize      Flash32k
#define  PageSize       Page128
#define  EEpromSize     EEprom1024
#define  PageByte       128
#define  NSHIFTPAGE     7
#define  INTVECREG      MCUCR
#define  PULLUPPORT     PORTE
#define  PULLUPPIN      0x01
#endif

#if defined(__AVR_ATmega6450__)
#define  DeviceID       Mega6450
#define  FlashSize      Flash64k
#define  PageSize       Page128
#define  EEpromSize     EEprom2048
#define  PageByte       256
#define  NSHIFTPAGE     8
#define  INTVECREG      MCUCR
#define  PULLUPPORT     PORTE
#define  PULLUPPIN      0x01
#endif

#if defined(__AVR_ATmega3290__)
#define  DeviceID       Mega3290
#define  FlashSize      Flash32k
#define  PageSize       Page128
#define  EEpromSize     EEprom1024
#define  PageByte       128
#define  NSHIFTPAGE     7
#define  INTVECREG      MCUCR
#define  PULLUPPORT     PORTE
#define  PULLUPPIN      0x01
#endif

#if defined(__AVR_ATmega6490__)
#define  DeviceID       Mega6490
#define  FlashSize      Flash64k
#define  PageSize       Page128
#define  EEpromSize     EEprom2048
#define  PageByte       256
#define  NSHIFTPAGE     8
#define  INTVECREG      MCUCR
#define  PULLUPPORT     PORTE
#define  PULLUPPIN      0x01
#endif

#if defined(__AVR_ATmega406__)
#define  DeviceID       Mega406
#define  FlashSize      Flash40k
#define  PageSize       Page128
#define  EEpromSize     EEprom512
#define  PageByte       128
#define  NSHIFTPAGE     7
#define  INTVECREG      MCUCR
#define  PULLUPPORT     PORTA
#define  PULLUPPIN      0x01
#endif

#if defined(__AVR_ATmega640__)
#define  DeviceID       Mega640
#define  FlashSize      Flash64k
#define  PageSize       Page256
#define  EEpromSize     EEprom4096
#define  PageByte       256
#define  NSHIFTPAGE     8
#define  INTVECREG      MCUCR
#if (UART == 0)
#define PULLUPPORT      PORTE
#define PULLUPPIN       0x01
#endif

#if (UART == 1)
#define PULLUPPORT      PORTD
#define PULLUPPIN       0x04
#endif

#if (UART == 2)
#define PULLUPPORT      PORTH
#define PULLUPPIN       0x01
#endif


#if (UART == 3)
#define PULLUPPORT      PORTJ
#define PULLUPPIN       0x01
#endif
#endif


#if defined(__AVR_ATmega1280__)
#define  DeviceID       Mega1280
#define  FlashSize      Flash128k
#define  PageSize       Page256
#define  EEpromSize     EEprom4096
#define  PageByte       256
#define  NSHIFTPAGE     8
#define  INTVECREG      MCUCR

#if (UART == 0)
#define PULLUPPORT      PORTE
#define PULLUPPIN       0x01
#endif

#if (UART == 1)
#define PULLUPPORT      PORTD
#define PULLUPPIN       0x04
#endif


#if (UART == 2)
#define PULLUPPORT      PORTH
#define PULLUPPIN       0x01
#endif


#if (UART == 3)
#define PULLUPPORT      PORTJ
#define PULLUPPIN       0x01
#endif
#endif

#if defined(__AVR_ATmega2560__)
#define  DeviceID       Mega2560
#define  FlashSize      Flash256k
#define  PageSize       Page256
#define  EEpromSize     EEprom4096
#define  PageByte       256
#define  NSHIFTPAGE     8
#define  INTVECREG      MCUCR

#if (UART == 0)
#define PULLUPPORT      PORTE
#define PULLUPPIN       0x01
#endif

#if (UART == 1)
#define PULLUPPORT      PORTD
#define PULLUPPIN       0x04
#endif


#if (UART == 2)
#define PULLUPPORT      PORTH
#define PULLUPPIN       0x01
#endif


#if (UART == 3)
#define PULLUPPORT      PORTJ
#define PULLUPPIN       0x01
#endif
#endif

#if defined (_AVR_IOMX8_H_)
#define FE         FE0
#define UPE        UPE0
#define RXC        RXC0
#endif

// Serial Port definition



#ifndef UDR
 #define PE UPE

#if (UART == 0)
#define  UCSRA          UCSR0A
#define  UCSRB          UCSR0B
#define  UCSRC          UCSR0C
#define  UBRRL          UBRR0L
#define  UBRRH          UBRR0H
#define  UDR            UDR0
#endif

#if (UART == 1)
#define  UCSRA          UCSR1A
#define  UCSRB          UCSR1B
#define  UCSRC          UCSR1C
#define  UBRRL          UBRR1L
#define  UBRRH          UBRR1H
#define  UDR            UDR1
#endif

#if (UART == 2)
#define  UCSRA          UCSR2A
#define  UCSRB          UCSR2B
#define  UCSRC          UCSR2C
#define  UBRRL          UBRR2L
#define  UBRRH          UBRR2H
#define  UDR            UDR2
#endif

#if (UART == 3)
#define  UCSRA          UCSR3A
#define  UCSRB          UCSR3B
#define  UCSRC          UCSR3C
#define  UBRRL          UBRR3L
#define  UBRRH          UBRR3H
#define  UDR            UDR3
#endif
#endif

#define  FALSE          (0)
#define  TRUE           (!TRUE)

/*****************************************************************************/
/*                          I N C L U D E                                    */
/*****************************************************************************/
#include "assembly.h"

/*****************************************************************************/
/*                        P R O T O T Y P E                                  */
/*****************************************************************************/
#define IsChar() (UCSRA & 0x80)
#define RxChar() (UDR)

static inline void FlashLoad(void);
#ifdef EEPROM
static inline void EEpromLoad(void);
static inline unsigned char EEpromWriteAndVerify(unsigned int location, unsigned char byte);
#endif
static inline void SendDeviceID(void);
static inline char GetPage(void);
static inline void WriteFlash(unsigned int PageAddress);
static inline char CheckFlash(unsigned int PageAddress);
#ifdef LOCKBIT
static inline void LockBit(void);
#endif
unsigned char getRxChar(void);
void TxChar(unsigned char ch);
static void Wait(void);
#ifdef HOMENET
static char HomeNet(void);
#endif
int main(void);

/*****************************************************************************/
/*                G L O B A L    V A R I A B L E S                           */
/*****************************************************************************/
unsigned char PageBuffer[PageByte];

/*****************************************************************************/
/* Flash Programing Code                                                     */
/*****************************************************************************/
/*****************************************************************************/

int main(void)
{
//  PULLUPPORT = PULLUPPIN;           // Pull up on RX line

  UCSRB = 0x00;                     //disable while setting baud rate
  UCSRA = 0x00;
  UCSRC = 0x86;                     // Asyn,NoParity,1StopBit,8Bit,
  UBRRH = 0x00;                     //set baud rate hi
  UCSRB = 0x18;                     // Rx enable Tx Enable

#ifdef RS485DDR
  RS485DDR |= RS485TXE;             // RS485 Tranceiver switch pin as output
  RS485PORT &= ~RS485TXE;           // RS485 in Rx mode
#endif

// Fr c't Microkontroller im Lan Projekt
#if defined (CT_PROJ)
 #ifndef BAUDRATE
    #error "You need to define the BAUDRATE for the c't microcontroler project"
 #endif
    DDRB  |= 0x07;  // Multiplexleitungen seriell PB0-2 als Ausgang
 #if defined (DEST_XPORT)
    PORTB &= 0xFC;  // 0bxxxxxx00                   // PB0=0; PB1=0  ==> Z=Z0 Y=Y0
    PORTB |= 0x04;  // 0b00000100                   // PB2=1 ==> (X=X1)
 #else //defined (DEST_COM)
    PORTB |=  0x03; // 0bxxxxxx11                   // PB0=1; PB1=1  ==> Z=Z1 Y=Y1
    PORTB &=  0xFB; // 0bxxxxx0xx                   // PB2=0 ==> (X=X0) (eigentlich don't care)
 #endif
#endif

  // Auto-find baudrate if BAUDRATE is undefine

#ifdef BAUDRATE
  UBRRL = (F_CPU/16/BAUDRATE)-1;
  RxChar();
  TxChar(START_CH);
  Wait();
  // ADDED wb :: (UCSRA & (_BV(FE)|_BV(UPE))) == 0 &&   :: Framing Error +/* Parity Error is BAD
  if ((UCSRA & (_BV(FE)|_BV(PE)|_BV(RXC))) != _BV(RXC) || RxChar() != START_CH_RET)
  {
    goto finish;
  }

#else
  unsigned char i;

  for (i = 0; i < 255; ++i)
  {
    UBRRL = i;
    RxChar();
    TxChar(START_CH);
    Wait();
    // :: Framing Error and/or Parity Error is BAD
    if ((UCSRA & (_BV(FE)|_BV(PE)|_BV(RXC))) == _BV(RXC) && RxChar() == START_CH_RET)
        break;
  }
  if (i == 255 )
  {
    goto finish;
  }
#endif


  // Auto-find OSCCAL
#ifdef AUTOOSCCAL
#ifdef BAUDRATE
  unsigned char i;
#endif
  unsigned char Min = 0xff;
  unsigned char Max = 0x00;
  EEAR = 0;
  EECR |= 0x01;                       // Set READ strobe
  if (EEDR == 0xff)
  {
    for (i=0;i<255;i++)
    {
      OSCCAL = i;
      TxChar(0x55);
      Wait();
      if (IsChar())
      {
        if (RxChar() == 0x55)
        {
          if (Min > i) Min = i;
          if (Max < i) Max = i;
        }
      }
    }
    OSCCAL = ((Max-Min)>>1)+Min;

    EEAR = 0;
    EEDR = OSCCAL;
    EECR |= 0x04;                       // Set MASTER WRITE enable
    EECR |= 0x02;                       // Set WRITE strobe
  }
  else
  {
    OSCCAL = EEDR;
  }
#endif

  SendDeviceID();

  FlashLoad();

#ifdef EEPROM
  EEpromLoad();
#endif

#ifdef LOCKBIT
  LockBit();
#endif

finish:
#if defined (RAMPZ)
  RAMPZ = 0;
#endif
#if defined (INTVECREG)
  INTVECREG = 0x01;	   	 // Enable interrupt vector select
  INTVECREG = 0x00;	   	 // Move interrupt vector to flash
#endif
  // Clean up: Stop the USART - TxD pin released.
  UCSRB = 0;                     // Rx disable Tx disable

  asm volatile ("push r1\n\t"                // Run application code
                "push r1\n\t"
                "ret\n\t" ::);
  /*
   * Never return to stop GCC to include a reference to the exit code.
   * Actually we will never reach this point, but the compiler doesn't
   * understand the assembly statement above.
   */
  for (;;);
}

/*****************************************************************************/

static inline char GetPage(void)
{
    unsigned int i = 0;
    unsigned char LocalCheckSum = 0;

    while (i < PageByte)
	{
	    LocalCheckSum += (PageBuffer[i++] = getRxChar());
   	}
    return (LocalCheckSum == getRxChar());
}

/*****************************************************************************/

static inline void WriteFlash(unsigned int PageAddress)
{
    unsigned int i = 0;
    unsigned int *ip = (unsigned int *)&PageBuffer[0];

    while (i < PageByte)
   	{
   	    fill_temp_buffer(*ip++, i);    //call asm routine.
   	    i += 2;
   	}

    write_page(PageAddress,0x03);       //Perform page ERASE
    write_page(PageAddress,0x05);       //Perform page write
    enableRWW();
}

/*****************************************************************************/

static inline char CheckFlash(unsigned int PageAddress)
{
    unsigned int i = 0;
    unsigned int *ip = (unsigned int *)&PageBuffer[0];

    while (i < PageByte)
    {
	    if (*ip++ != read_program_memory(PageAddress + i,0x00))
	        return 0;
	    i += 2;
    }
    return 1;
}

/*****************************************************************************/

static inline void FlashLoad(void)
{
unsigned int RealPageAddress;
unsigned int PageAddress;

    unsigned char ret;
#ifdef MEGALOAD_NET
    TxChar('>');
#endif
    TxChar('!');
    while (1)
	{
	    ret = '@';
        RealPageAddress = (unsigned int)((getRxChar() << 8) + getRxChar());
	    if (RealPageAddress == 0xffff)
	        return;
#if defined (RAMPZ)
        RAMPZ = (RealPageAddress & 0xFF00) ? 1 : 0;
#endif
        PageAddress = RealPageAddress << NSHIFTPAGE;
	    if (GetPage())
		{
		    WriteFlash(PageAddress);
		    if (CheckFlash(PageAddress))
		    {
		        ret = '!';
		    }
		}
        TxChar(ret);
	 }
}

/*****************************************************************************/
/* EEprom Programing Code                                                    */
/*****************************************************************************/
#ifdef EEPROM
static inline unsigned char EEpromWriteAndVerify( unsigned int location, unsigned char byte )
{
//    while (EECR & 0x02)    // Wait until any earlier write is done
//        ;
    EEAR = location;
    EEDR = byte;
    EECR |= 0x04;          // Set MASTER WRITE enable
    EECR |= 0x02;          // Set WRITE strobe
    while (EECR & 0x02)
        ;
    EECR |= 0x01;                       // Set READ strobe
    return EEDR == byte;
}

/*****************************************************************************/

static inline void EEpromLoad(void)
{
    union {
      unsigned int ba;
      struct {
        unsigned char baLow;
        unsigned char baHigh;
      };
    } ByteAddress;
    unsigned char Data;
    unsigned char LocalCheckSum;
    unsigned char res;

    TxChar(')');
    TxChar('!');
    while (1)
    {
        res = '@';
        LocalCheckSum  = ByteAddress.baHigh = getRxChar();
        LocalCheckSum += (ByteAddress.baLow = getRxChar());
        if (ByteAddress.ba == 0xffff)
            return;
        LocalCheckSum += (Data = getRxChar());
        if ((LocalCheckSum == getRxChar())
            && (EEpromWriteAndVerify(ByteAddress.ba, Data)))
        {
            res = '!';
        }
        TxChar(res);
    }
}
#endif


/*****************************************************************************/
/* LockBit Code                                                              */
/*****************************************************************************/
#ifdef LOCKBIT
static inline void LockBit(void)
{
    unsigned char Byte;

    TxChar('%');

    Byte = ~getRxChar();

    if (Byte == getRxChar())
//        boot_lock_bits_set(Byte);
        write_lock_bits(Byte);
}
#endif
/*****************************************************************************/
/* Serial Port Code                                                          */
/*****************************************************************************/
/*****************************************************************************/
unsigned char getRxChar(void)
{
    while(!IsChar())
        ;
    return UDR;
}
/*****************************************************************************/

void TxChar(unsigned char ch)
{
    while(!(UCSRA & 0x20))   // wait for empty transmit buffer
        ;
#ifndef RS485DDR
    UDR = ch;     	 		 // write char
#endif

#ifdef RS485DDR
    RS485PORT |= RS485TXE;	 // RS485 in TX mode
    UDR = ch;     	 		 // write char
    while(!(UCSRA & 0x40))   // Wait for char to be cue off
        ;
    UCSRA |= 0x40;			 // Clear flag
    RS485PORT &= ~RS485TXE;	 // RS485 in RX mode
#endif
}

/*****************************************************************************/
/* Helpers Code                                                              */
/*****************************************************************************/
static void Wait(void)
{
#if (F_CPU < 8000000)
    int i = 0;
#else
    int i = -32000;
#endif
    for (;i<32000;i++)
        asm volatile ("nop" ::);
}

/*****************************************************************************/

static inline void SendDeviceID(void)
{
  TxChar(DeviceID);
  TxChar(FlashSize);
  TxChar(BootSize);
  TxChar(PageSize);
  TxChar(EEpromSize);
  RxChar();
}

/*****************************************************************************/

/*****************************************************************************/
/*****************************************************************************/
#ifdef HOMENET
char HomeNet(void)
{
    unsigned char i;
    unsigned char Exec = 0;
    unsigned char Boot = 0;
    unsigned char Address;

    EEAR = 0x01;
    EECR |= 0x01;                       // Set READ strobe
    Address = EEDR;                     // Return byte

    while(1)
	{
		i = getRxChar();

		if (i == '>')
		    Exec++;
		else
		    Exec = 0;

		if (i == '+')
		    Boot++;
		else
		    Boot = 0;

		if (Exec >= 3)
		    return 1;
		if (Boot >= 5)
		{
		    if (getRxChar() != Address)
		        return 1;
		    return 0;
	    }
    }
    return 1;
}
#endif

